# Copyright 2005-2012 Canonical Limited.  All rights reserved.

from datetime import datetime
from cStringIO import StringIO
from pydoc import TextDoc
import json

import pycurl

from twisted.trial.unittest import TestCase

from landscape_api.base import (
    parse, run_query, API, main, HTTPError, _build_exceptions, APIError,
    LATEST_VERSION, api_factory, _API)
from landscape_api.base.errors import (
    InvalidActivityStatusChangeError, UnknownComputerError,
    UnknownAPTSourceError, MultiError, GPGKeyImportError,
    AuthError)
from mocker import MockerTestCase


TEST_SCHEMA = {
    "GetFoo": {
        "2011-08-01": {
            "parameters": [
                {"name": "id", "type": "integer"}
            ]
        }
    },
    "MethodWithOptional": {
        "2011-08-01": {
            "parameters": [
                {"name": "id", "type": "integer",
                 "optional": True, "default": 5}
            ]
        }
    },
    "MethodWithList": {
        "2011-08-01": {
            "parameters": [
                {"name": "listy", "type": "list", "item": {"type": "integer"}}
            ]
        }
    },
    "MethodWithStructure": {
        "2011-08-01": {
            "parameters": [
                {
                    "name": "struct",
                    "type": "structure",
                    "fields": {
                        "field": {"type": "unicode"},
                        "value": {"type": "integer"}
                    }
                }
            ]
        }
    },
    "MethodWithUnicode": {
        "2011-08-01": {
            "parameters": [
                {"name": "uni", "type": "unicode"}
            ]
         }
    },
    "MethodWithRawString": {
        "2011-08-01": {
            "parameters": [
                {"name": "raw", "type": "raw string"}
            ]
        }
    },
    "MethodWithDate": {
        "2011-08-01": {
            "parameters": [
                {"name": "date", "type": "date"}
            ]
        }
    },
    "MethodWithEnum": {
        "2011-08-01": {
            "parameters": [
                {"name": "enum", "type": "enum"}
            ]
        }
    },
    "MethodWithBoolean": {
        "2011-08-01": {
            "parameters": [
                {"name": "booly", "type": "boolean"},
                {"name": "other", "type": "boolean"}
            ]
        }
    },
    "MethodWithOptionalStructureField": {
        "2011-08-01": {
            "parameters": [
                {"name": "struct",
                 "type": "structure",
                 "fields": {
                        "a": {"type": "integer", "optional": True}
                    }
                }
            ]
        }
    },
    "MethodWithOptionalBoolean": {
        "2011-08-01": {
            "parameters": [
                {"name": "deffalse", "type": "boolean",
                 "optional": True, "default": False},
                {"name": "deftrue", "type": "boolean",
                 "optional": True, "default": True}
            ]
        }
    },
    "HelpyMethod": {
        "2011-08-01": {
            "doc": "This method has help.",
            "parameters": [
                {"name": "cheese", "type": "boolean",
                 "doc": "This parameter has help."}
            ]
        }
    }
}

for action_name, version_handlers in TEST_SCHEMA.iteritems():
    for version, handler in version_handlers.iteritems():
        if "doc" not in handler:
            handler["doc"] = "Documentation exists!"
        for parameter in handler["parameters"]:
            if "doc" not in parameter:
                parameter["doc"] = "Parameter documentation exists!"


class RawTextDoc(TextDoc):
    """A C{pydoc.TextDoc} class without bold capability."""

    def bold(self, text):
        return text


class BaseTestCase(MockerTestCase):

    def test_parse(self):
        """
        L{parse} takes an URI and returns the parsed host/port/path from it.
        """
        self.assertEqual(
            ("localhost", None, "/"), parse("http://localhost/"))
        self.assertEqual(
            ("localhost", 123, "/foo"), parse("http://localhost:123/foo"))
        self.assertEqual(
            ("localhost", None, "/"), parse("http://localhost:aa/"))

    def test_run_query(self):
        """
        L{run_query} does the signing with the given credentials, and extracts
        the host properly to pass the C{Host} header field.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings", {"arg1": "Value1"},
                         "http://localhost/")
        self.assertEqual("data", data)

    def test_run_query_special_exception(self):
        """
        L{run_query} is able to translate known errors from the API into
        exceptions.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "InvalidActivityStatusChange",
                            "message": "oops"})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        activity_error = self.assertRaises(
            InvalidActivityStatusChangeError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(400, activity_error.code)
        self.assertEqual("oops", activity_error.error_message)
        self.assertEqual("InvalidActivityStatusChange",
                         activity_error.error_code)

    def test_run_query_auth_error(self):
        """
        L{AuthError} is raised when an error is returned with error code set to
        "AuthError". This works even though the HTTP code is != 400.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "AuthError",
                            "message": "oops"})
        self.mocker.throw(HTTPError(403, error))
        self.mocker.replay()

        auth_error = self.assertRaises(
            AuthError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(403, auth_error.code)
        self.assertEqual("oops", auth_error.error_message)
        self.assertEqual("AuthError", auth_error.error_code)

    def test_run_query_error_code_ending_in_error(self):
        """
        If an error code ends in "Error", it is looked up directly instead of
        appending "Error" to the name.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "GPGKeyImportError", "message": "oops"})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        self.assertRaises(
            GPGKeyImportError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")

    def test_run_query_unknown_exception(self):
        """
        If the API error returned is not known by the client, L{run_query}
        raises L{HTTPError} as usual.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "ActivityFailure", "message": "oops"})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        http_error = self.assertRaises(
            HTTPError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(400, http_error.code)
        self.assertEqual(error, http_error.message)

    def test_run_query_multi_error(self):
        """
        If the API error returned is a MultiError, then a L{MultiError} is
        constructed containing all of the sub-errors and raised.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "MultiError",
                            "message": "Multiple errors",
                            "errors": [{"error": "UnknownComputer",
                                        "message": "Foo"},
                                       {"error": "UnknownAPTSource",
                                        "message": "Bar"}]
                            })
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        multi_error = self.assertRaises(
            MultiError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(400, multi_error.code)
        self.assertEqual(2, len(multi_error.errors))
        self.assertTrue(isinstance(multi_error.errors[0],
                                   UnknownComputerError))
        self.assertTrue(isinstance(multi_error.errors[1],
                                   UnknownAPTSourceError))

    def test_multi_error_with_error_code_ending_in_error(self):
        """
        L{MultiError} uses L{_get_error_code_name} to look up the error. i.e.,
        errors ending in "Error" do not try to raise an exception named
        "FooErrorError".
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))

        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "MultiError",
                            "message": "Multiple errors",
                            "errors": [{"error": "GPGKeyImportError",
                                        "message": "Foo"}]})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        multi_error = self.assertRaises(
            MultiError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")

        self.assertTrue(isinstance(multi_error.errors[0], GPGKeyImportError))

    def test_unknown_multi_error(self):
        """
        If a sub-error doesn't have a predefined exception class registered,
        then it will be a generic instance of L{APIError}.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "MultiError",
                            "message": "Multiple errors",
                            "errors": [{"error": "FooBar", "message": "Foo"}]})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        multi_error = self.assertRaises(
            MultiError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(400, multi_error.code)
        self.assertEqual(1, len(multi_error.errors))
        self.assertTrue(isinstance(multi_error.errors[0], APIError))
        self.assertEqual("FooBar", multi_error.errors[0].error_code)
        self.assertEqual("Foo", multi_error.errors[0].error_message)

    def test_uncoded_multi_error(self):
        """
        MultiErrors containing errors without codes are parsed correctly.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        error = json.dumps({"error": "MultiError",
                            "message": "Multiple errors",
                            "errors": [{"error": None, "message": "Foo"}]})
        self.mocker.throw(HTTPError(400, error))
        self.mocker.replay()

        multi_error = self.assertRaises(
            MultiError, run_query, "access", "secret",
            "GetThings", {"arg1": "Value1"}, "http://localhost/")
        self.assertEqual(400, multi_error.code)
        self.assertEqual(1, len(multi_error.errors))
        self.assertTrue(isinstance(multi_error.errors[0], APIError))
        self.assertEqual(None, multi_error.errors[0].error_code)
        self.assertEqual("Foo", multi_error.errors[0].error_message)

    def test_run_query_with_port(self):
        """
        If the API call happens on a port different than the standard one,
        L{run_query} passes it in the C{Host} header.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=TfIKl45NFsSX3QTO4lhJb42nQEM9X1NYTnnZB/hhyHI%3D")
        fetch("https://localhost:1234/foo", body, {"Host": "localhost:1234"},
              cainfo=None)
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings", {"arg1": "Value1"},
                         "https://localhost:1234/foo")
        self.assertEqual("data", data)

    def test_run_query_without_path(self):
        """
        If the final C{/} is ommitted in the URI, C{run_query} automatically
        adds it for the signing to work properly.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings", {"arg1": "Value1"},
                         "http://localhost")
        self.assertEqual("data", data)

    def test_run_query_parameter_encoding(self):
        """
        L{run_query} encodes the parameters as per the AWS Query API
        Authentication documentation on how to create the signature.
        Hyphens, underscores, periods and tilde shouldn't be encoded,
        but slashes should.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&"
                "arg-_.~%2F=Value-_.~%2F&signature_method=HmacSHA256&"
                "signature_version=2&timestamp=2011-07-11T08%3A46%3A01Z&"
                "version=2011-08-01&signature="
                "TxXjdYi5O48m/v3fhR/GePILmdv80gL39HuX7ynqoRw%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings",
                         {"arg-_.~/": "Value-_.~/"}, "http://localhost/")
        self.assertEqual("data", data)

    def test_run_query_with_unicode_parameter(self):
        """
        L{run_query} also accepts unicode parameters.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&"
                "arg-_.~%2F%E2%82%AC=Value-_.~%2F%E2%82%AC&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=jUenQyo7LclhVo8LfrC0sDO6vX3LuxQdDXi4R5cCJCM%3D")
        fetch("http://localhost/", body, {"Host": "localhost"}, cainfo=None)
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings",
                         {u"arg-_.~/\u20ac": u"Value-_.~/\u20ac"},
                         "http://localhost/")
        self.assertEqual("data", data)

    def test_run_query_with_ssl_ca_file(self):
        """
        If we have an SSL CA certificate file to verify the server against,
        then this should be passed to fetch.
        """
        gmtime = self.mocker.replace("time.gmtime")
        gmtime()
        self.mocker.result((2011, 7, 11, 8, 46, 1, 0, 0, 0))
        fetch = self.mocker.replace(
            "landscape_api.base.fetch", passthrough=False)
        body = ("access_key_id=access&action=GetThings&arg1=Value1&"
                "signature_method=HmacSHA256&signature_version=2&"
                "timestamp=2011-07-11T08%3A46%3A01Z&version=2011-08-01&"
                "signature=QYlDFMTF/BpfF8ihYpKdShWh8d02manhauGOpKVPAos%3D")
        fetch("https://localhost/", body, {"Host": "localhost"},
              cainfo="/path/to/ssl")
        self.mocker.result("data")
        self.mocker.replay()
        data = run_query("access", "secret", "GetThings", {"arg1": "Value1"},
                         "https://localhost/", ssl_ca_file="/path/to/ssl")
        self.assertEqual("data", data)

    def test_query_return_object(self):
        """
        By default, API calls will return a python object as response.
        """
        api = API("http://localhost", "access", "secret")

        def run_query(self, *args):
            return json.dumps({"json": "is cool"})

        api._run_query = run_query

        self.assertEqual({"json": "is cool"},
                         api.run_query("DoStuff", {"foo": "some-query"}))

    def test_query_return_json(self):
        """
        If C{json} parameter is true, API calls return plain JSON string
        response.
        """
        api = API("http://localhost", "access", "secret", json=True)

        def run_query(self, *args):
            return json.dumps({"json": "is cool"})

        api._run_query = run_query

        self.assertEqual(json.dumps({"json": "is cool"}),
                         api.run_query("DoStuff", {"foo": "some-query"}))

    def test_call_arbitrary(self):
        """
        L{API.call_arbitrary} allows invoking an arbitrary API method with a
        dictionary of already-encoded arguments.
        """
        calls = []
        api = API("uri", "access", "secret")

        def run_query(*args):
            calls.append(args)
            return '{"foo": "bar"}'

        api._run_query = run_query
        result = api.call_arbitrary("GetComputers", "foo.bar=5")
        self.assertEqual({"foo": "bar"}, result)
        self.assertEqual(
            [('access', 'secret', 'GetComputers', 'foo.bar=5', 'uri', None)],
            calls)

    def test_call_arbitrary_honors_json(self):
        """
        If the C{json} argument to API is passed as True, L{API.call_arbitrary}
        returns json.
        """
        api = API("uri", "access", "secret", json=True)

        def run_query(*args):
            return '{"foo": "bar"}'

        api._run_query = run_query
        result = api.call_arbitrary("GetComputers", "foo.bar=5")
        self.assertEqual('{"foo": "bar"}', result)

    def test_duplicate_keyword_arguments(self):
        """
        If a positional argument for a parameter is passed along with a keyword
        argument for the same parameter, a L{TypeError} is raised, just as if
        it were a function defined with real positional arguments.
        """
        api = api_factory(TEST_SCHEMA)(
            "uri", "access", "secret", schema=TEST_SCHEMA)
        e = self.assertRaises(TypeError, api.get_foo, 1, id=2)
        self.assertEqual(
            "get_foo() got multiple values for keyword argument 'id'",
            str(e))

    def test_dont_replace_core_methods(self):
        """
        If a schema defines an action name that would result in overwriting a
        method of L{API}, then instantiating L{API} blows up.
        """
        schema = TEST_SCHEMA.copy()
        schema["Call"] = {LATEST_VERSION: {"parameters": [], "doc": "Call!"}}
        e = self.assertRaises(RuntimeError, api_factory, schema)
        self.assertIn(
            "Tried setting 'call' from schema but that method already exists",
            str(e))

    def test_pydoc_on_class(self):
        """
        Methods on L{API} classes are defined statically and their
        documentation contains argument names.
        """
        api_class = api_factory(TEST_SCHEMA)
        doc = RawTextDoc().document(api_class.get_foo)
        self.assertIn("get_foo(self, id)", doc)

    def test_pydoc_argument_method_help(self):
        """
        API methods defined on L{API} have their help showing the schema help
        text, if any.
        """
        api_class = api_factory(TEST_SCHEMA)
        doc = RawTextDoc().document(api_class.helpy_method)
        self.assertIn("This method has help.", doc)

    def test_pydoc_default_argument(self):
        """
        API methods defined on L{API} have the default values of their argument
        listed when using pydoc.
        """
        api_class = api_factory(TEST_SCHEMA)
        doc = RawTextDoc().document(api_class.method_with_optional)
        self.assertIn("method_with_optional(self, id=5)", doc)
        doc = RawTextDoc().document(api_class.method_with_optional_boolean)
        self.assertIn(
            "method_with_optional_boolean(self, deffalse=False, deftrue=True)",
            doc)

    def test_pydoc_includes_parameters(self):
        """
        API methods defined on L{API} have parameter information in their
        documentation.
        """
        api_class = api_factory(TEST_SCHEMA)
        doc = RawTextDoc().document(api_class.method_with_optional)
        self.assertIn("method_with_optional(self, id=5)", doc)
        doc = RawTextDoc().document(api_class.method_with_optional_boolean)
        self.assertIn("@param deffalse: Parameter documentation exists!", doc)
        self.assertIn("@type deffalse: boolean", doc)


class MainTests(MockerTestCase):

    def setUp(self):
        super(MainTests, self).setUp()
        self.stdout = StringIO()
        self.stderr = StringIO()
        self.exits = []
        self.exit = self.exits.append
        self.calls = []
        self.addCleanup(setattr, _API, "_run_query", staticmethod(run_query))
        _API._run_query = self.run_query
        self.env = {"LANDSCAPE_API_KEY": "key",
                    "LANDSCAPE_API_SECRET": "secret",
                    "LANDSCAPE_API_URI": "uri"}

    def run_query(self, *args):
        self.calls.append(args)
        return json.dumps("data")

    def test_no_arguments(self):
        """
        When no action is specified, an error is reported.
        """
        self.call_main(["--key", "k", "--secret", "s", "--uri", "u"])
        self.assertIn("Please specify an action.", self.stderr.getvalue())

    def test_help(self):
        """
        The C{--help} argument prints command usage, and exit with 0.
        """
        main(["--help"], self.stdout, self.stderr, self.exit, {})
        self.assertIn("usage:", self.stdout.getvalue())
        self.assertNotIn("@param", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([0], self.exits)

    def test_action_help(self):
        """
        The C{help} argument prints command usage, and exit with 0.
        """
        self.call_main(["help"])
        self.assertIn("usage:", self.stdout.getvalue())
        self.assertNotIn("@param", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([0], self.exits)

    def test_short_help(self):
        """
        The C{-h} argument prints command usage, and exit with 0.
        """
        main(["-h"], self.stdout, self.stderr, self.exit, {})
        self.assertIn("usage:", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([0], self.exits)

    def test_no_access_key(self):
        """
        If the access key is not specified, an error message is given.
        """
        main(["--secret", "s", "--uri", "u", "get-computers"],
             self.stdout, self.stderr, self.exit, {})
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("Access key not specified.",
                      self.stderr.getvalue())
        self.assertEqual([1], self.exits)

    def test_no_secret_key(self):
        """
        If the secret key is not specified, an error message is given.
        """
        main(["--key", "k", "--uri", "u", "get-computers"],
             self.stdout, self.stderr, self.exit, {})
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("Secret key not specified.",
                      self.stderr.getvalue())
        self.assertEqual([1], self.exits)

    def test_no_uri(self):
        """
        If the URI is not specified, an error message is given.
        """
        main(["--key", "k", "--secret", "s", "get-computers"],
             self.stdout, self.stderr, self.exit, {})
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("URI not specified.",
                      self.stderr.getvalue())
        self.assertEqual([1], self.exits)

    def test_use_environment(self):
        """
        The script falls back to use the environment values if they exist.
        """
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual([("key", "secret", "GetComputers", {}, "uri", None)],
                         self.calls)
        self.assertEqual("u'data'\n", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([], self.exits)

    def test_use_parameters(self):
        """
        The parameters to the command line overrides the values of the
        environment.
        """
        main(["--key", "key2", "--secret", "secret2", "--uri", "uri2",
              "get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual([("key2", "secret2", "GetComputers", {}, "uri2",
                           None)], self.calls)
        self.assertEqual("u'data'\n", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([], self.exits)

    def test_missing_argument_does_not_raise(self):
        """
        If required arguments are missing from a command, a nice error message
        should be displayed instead of a traceback.
        """
        main(["install-packages"], self.stdout,
             self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("usage:", self.stderr.getvalue())
        self.assertIn("error: too few arguments",
                      self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_positional_argument(self):
        """
        A specific error message is displayed if positional arguments are
        passed to a command which doesn't accept a C{query} parameter.
        """
        main(["get-scripts", "foo"], self.stdout, self.stderr, self.exit,
             self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("usage:", self.stderr.getvalue())
        self.assertIn("error: unrecognized arguments: foo",
                      self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_query_argument(self):
        """
        If the C{--query} option is used on a command which doesn't accept the
        parameter, the normal error message is used.
        """
        main(["get-scripts", "--query", "foo"], self.stdout, self.stderr,
             self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("usage:", self.stderr.getvalue())
        self.assertIn("unrecognized arguments: --query foo",
                      self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_argument(self):
        """
        An error message is displayed if an unknown parameter is passed to a
        command.
        """
        main(["get-computers", "--unknown", "bar"], self.stdout,
             self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("usage:", self.stderr.getvalue())
        self.assertIn("error: unrecognized arguments: --unknown bar",
                      self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_action(self):
        """
        An error message is displayed if an unknown action is passed to the
        script.
        """
        main(["unknown-action"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("Unknown action: unknown-action",
                      self.stderr.getvalue())
        self.assertEqual([1], self.exits)

    def test_action_not_specified(self):
        """
        If the action is not specified to the script, a help message is
        displayed.
        """
        main([], self.stdout, self.stderr, self.exit, self.env)
        self.assertIn("usage:", self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([0], self.exits)

    def test_missing_option_argument(self):
        """
        If an option doesn't have a value, an error message is displayed.
        """
        main(["get-computers", "--limit", "3", "--offset"], self.stdout,
             self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertIn("usage:", self.stderr.getvalue())
        self.assertIn("error: argument --offset: expected one argument",
                      self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_multi_word_argument(self):
        """
        If an optional argument contains multiple words, separated by a
        dash, the dash is converted to an underscore so that it can be
        passed as a keyword parameter to the action method.
        """
        main(["get-packages", "--search", "foo", "query"],
             self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual(
            [("key", "secret", "GetPackages",
             {"query": "query", "search": "foo"}, "uri", None)],
            self.calls)

    def test_json_output(self):
        """
        The C{--json} makes the script output the JSON data directly.
        """
        main(["--json", "get-computers"], self.stdout, self.stderr, self.exit,
             self.env)
        self.assertEqual([("key", "secret", "GetComputers", {}, "uri", None)],
                         self.calls)
        self.assertEqual('%s\n' % json.dumps("data"), self.stdout.getvalue())
        self.assertEqual("", self.stderr.getvalue())
        self.assertEqual([], self.exits)

    def test_known_http_error(self):
        """
        If a HTTP C{400} error is received, the body is deserialized and an
        error message is extracted from it.
        """
        def run_query(*args):
            raise HTTPError(400, json.dumps({"message": "Woops!",
                                             "error": "InvalidQuery"}))
        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot server error:\nStatus: 400\n"
                         "Error code: InvalidQuery\nError message: Woops!\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_multi_error(self):
        """
        If a MultiError is received, each individual error is printed.
        """
        def run_query(*args):
            raise MultiError(400,
                            json.dumps({"message": "MultiWoops!",
                                        "error": "MultiError",
                                        "errors": [{"error": "InvalidQuery",
                                                    "message": "Woops!"},
                                                   {"error": None,
                                                    "message": "Unknown"}]}))
        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot server error:\nStatus: 400\n"
                         "Error code: MultiError\nError message: MultiWoops!\n"
                         "Error code: InvalidQuery\nError message: Woops!\n"
                         "Error code: None\nError message: Unknown\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_http_error_code(self):
        """
        If a HTTP C{400} error is received with an unknown error code, the body
        is deserialized and an error message is extracted from it.
        """
        def run_query(*args):
            raise HTTPError(400, json.dumps({"message": "Woops!",
                                             "error": "Unknown"}))
        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot server error:\nStatus: 400\n"
                         "Error code: Unknown\nError message: Woops!\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unicode_error(self):
        """
        If the error returned has a unicode message, it's correctly displayed
        in UTF-8.
        """
        message = u"prova\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}o"

        def run_query(*args):
            raise HTTPError(400, json.dumps({"message": message,
                                             "error": u"ErrorCode"}))

        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot server error:\nStatus: 400\n"
                         "Error code: ErrorCode\nError message: "
                         "prova\xc4\xb5o\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_non_400_http_error(self):
        """
        HTTP error responses with codes other than 400 are displayed the same
        way as 400 errors, as long as they still have JSON bodies.
        """
        def run_query(*args):
            raise HTTPError(401, json.dumps({"message": "Woops!",
                                             "error": "InvalidQuery"}))
        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot server error:\nStatus: 401\n"
                         "Error code: InvalidQuery\nError message: Woops!\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_unknown_http_error(self):
        """
        If a HTTP error with a non-JSON body is received, the code and the
        body of the error are displayed.
        """
        def run_query(*args):
            raise HTTPError(500, "Failure!")
        _API._run_query = run_query
        main(["get-computers"], self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual("", self.stdout.getvalue())
        self.assertEqual("\nGot unexpected server error:\nStatus: 500\n"
                         "Error message: Failure!\n",
                         self.stderr.getvalue())
        self.assertEqual([2], self.exits)

    def test_ssl_ca_file(self):
        """
        The C{--ssl-ca-file} option provides an SSL CA certificate to validate
        the server.
        """
        main(["--ssl-ca-file", "/path/to/ssl", "get-computers"],
             self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual([("key", "secret", "GetComputers", {}, "uri",
                           "/path/to/ssl")], self.calls)

    def test_ssl_ca_file_with_environment_variable(self):
        """
        If C{--ssl-ca-file} is not provided, the client defaults to the value
        of the LANDSCAPE_API_SSL_CA_FILE environment variable.
        """
        self.env["LANDSCAPE_API_SSL_CA_FILE"] = "/path/to/env/ssl"
        main(["get-computers"],
             self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual([("key", "secret", "GetComputers", {}, "uri",
                           "/path/to/env/ssl")], self.calls)

    def test_call_action(self):
        """
        Using an action of C{call} allows invoking arbitrary API methods.
        """
        main(["call", "GetComputers", "limit=1"],
             self.stdout, self.stderr, self.exit, self.env)
        self.assertEqual([("key", "secret", "GetComputers", {"limit": "1"},
                           "uri", None)], self.calls)

    def assertCall(self, method, arguments):
        """
        Check what the last call made was.
        """
        if not self.calls:
            self.fail("There have been no calls. Perhaps this will help:\n"
                      "stdout:\n"
                      + self.stdout.getvalue() +
                      "stderr:\n"
                      + self.stderr.getvalue())
        self.assertEqual(method, self.calls[-1][2])
        self.assertEqual(arguments, self.calls[-1][3])

    def call_main(self, args, schema=None):
        """
        Call the L{main} function with faked out stdout, stderr, exit function,
        and environment.
        """
        kwargs = {}
        if schema is not None:
            kwargs["schema"] = schema
        return main(args, self.stdout, self.stderr, self.exit, self.env,
                    **kwargs)

    def test_parse_int(self):
        """
        Ints can be parsed.
        """
        self.call_main(["get-foo", "5"], TEST_SCHEMA)
        self.assertCall("GetFoo", {"id": "5"})

    def test_parse_list(self):
        """
        comma-delimited lists are parsed as lists.
        """
        self.call_main(["method-with-list", "5,6"], TEST_SCHEMA)
        self.assertCall("MethodWithList", {"listy.1": "5", "listy.2": "6"})

    def test_parse_unicode(self):
        """
        Unicode parameters are parsed as utf-8.
        """
        val = u"foo\N{HIRAGANA LETTER A}".encode("utf-8")
        self.call_main(["method-with-unicode", val], TEST_SCHEMA)
        self.assertCall("MethodWithUnicode", {"uni": val})

    def test_parse_raw_string(self):
        """
        Raw strings are passed through without changing the encoding.
        """
        # Let's get a value that's not decodable as utf-8... though this is
        # exceedingly uncommon
        val = "\xe3\xe3\xe3"
        self.call_main(["method-with-raw-string", val], TEST_SCHEMA)
        self.assertCall("MethodWithRawString", {"raw": val})

    def test_parse_date(self):
        """
        Dates are passed through.
        """
        self.call_main(["method-with-date", "2001-01-01"], TEST_SCHEMA)
        self.assertCall("MethodWithDate", {"date": "2001-01-01"})

    def test_parse_enum(self):
        """
        Enums are passed through.
        """
        # We don't do anything special with enums yet, though we could
        # theoretically limit them to the available options on the client.
        self.call_main(["method-with-enum", "foobar"], TEST_SCHEMA)
        self.assertCall("MethodWithEnum", {"enum": "foobar"})

    def test_parse_boolean(self):
        """
        Lowercase true/false values are parsed as booleans.
        """
        self.call_main(["method-with-boolean", "true", "false"], TEST_SCHEMA)
        self.assertCall("MethodWithBoolean",
                        {"booly": "true", "other": "false"})

    def test_parse_error_int(self):
        """
        Trying to pass a non-integer to an integer parameter results in an
        error.
        """
        self.call_main(["get-foo", "abc"], TEST_SCHEMA)
        self.assertIn("Couldn't parse value 'abc' as integer",
                      self.stderr.getvalue())
        self.assertEqual(self.exits, [1])

    def test_parse_error_list(self):
        """
        Errors while parsing elements of lists mention the type of the element,
        not list.
        """
        self.call_main(["method-with-list", "1,2,abc"], TEST_SCHEMA)
        self.assertIn("Couldn't parse value 'abc' as integer",
                      self.stderr.getvalue())
        self.assertEqual(self.exits, [1])

    def test_parse_optional_bool_default_false(self):
        """
        Optional booleans with a default of false get a command line flag that
        doesn't accept a value.
        """
        self.call_main(["method-with-optional-boolean", "--deffalse"],
                       TEST_SCHEMA)
        self.assertCall("MethodWithOptionalBoolean", {"deffalse": "true"})

    def test_parse_optional_bool_default_true(self):
        """
        Optional booleans with a default of true get a command line flag that
        doesn't accept a value and has "no-" at the beginning.
        """
        self.call_main(["method-with-optional-boolean", "--no-deftrue"],
                       TEST_SCHEMA)
        self.assertCall("MethodWithOptionalBoolean", {"deftrue": "false"})

    def test_help_call(self):
        """
        'help call' gives information about the 'call' action.
        """
        self.call_main(["help", "call"])
        self.assertIn("Call an arbitrary Landscape API action.",
                      self.stdout.getvalue())

    def test_dash_dash_help_call(self):
        """
        '--help call' gives information about the 'call' action.
        """
        self.call_main(["--help", "call"])
        self.assertIn("Call an arbitrary Landscape API action.",
                      self.stdout.getvalue())

    def test_call_dash_dash_help(self):
        """
        'call --help' gives information about the 'call' action.
        """
        self.call_main(["call", "--help"])
        self.assertIn("Call an arbitrary Landscape API action.",
                      self.stdout.getvalue())

    def test_help_action(self):
        """
        'help <action>' gives information about the action.
        """
        self.call_main(["help", "helpy-method"], TEST_SCHEMA)
        self.assertIn("This method has help.", self.stdout.getvalue())

    def test_dash_dash_help_action(self):
        """
        '--help <action>' gives information about the action.
        """
        self.call_main(["--help", "helpy-method"], TEST_SCHEMA)
        self.assertIn("This method has help.", self.stdout.getvalue())

    def test_action_dash_dash_help(self):
        """
        '<action> --help' gives information about the action.
        """
        self.call_main(["helpy-method", "--help"], TEST_SCHEMA)
        self.assertIn("This method has help.", self.stdout.getvalue())

    def test_overridden_apis_invokes_overridden_method(self):
        """
        L{API} can declare overridden actions that can replace parameters and
        invoke a different method.
        """
        self.addCleanup(setattr, API, "overridden_apis", API.overridden_apis)
        API.overridden_apis = {
            "GetFoo": {
                "replace_args": {"id": {"name": "id2", "doc": "heya",
                                        "type": "unicode"}},
                "method": "get_foo_with_id2"
                }
            }

        calls = []

        def get_foo_with_id2(self, id2):
            calls.append(id2)

        self.addCleanup(delattr, _API, "get_foo_with_id2")
        _API.get_foo_with_id2 = get_foo_with_id2

        self.call_main(["get-foo", "5"], TEST_SCHEMA)
        self.assertEqual(["5"], calls)

    def test_overridden_apis_no_duplicate_actions(self):
        """
        Overridden APIs replace the original action in the command line,
        instead of adding an additional one.
        """
        self.addCleanup(setattr, API, "overridden_apis", API.overridden_apis)
        API.overridden_apis = {
            "GetFoo": {
                "replace_args": {"id": {"name": "id2", "doc": "heya",
                                        "type": "unicode"}},
                "method": "get_foo_with_id2"
                }
            }
        self.call_main(["help"], TEST_SCHEMA)
        self.assertEqual(1, self.stdout.getvalue().count("get-foo"))

    def test_overridden_apis_doc(self):
        """
        Documentation in the overridden APIs are honored.
        """
        self.addCleanup(setattr, API, "overridden_apis", API.overridden_apis)
        API.overridden_apis = {
            "MethodWithBoolean": {
                "replace_args": {
                    "booly": {"name": "booly-file", "doc": "Booly File",
                              "type": "unicode"}
                    },
                "method": "method_with_boolean_override",
                "doc": "Overridden action documentation."
                }
            }

        def method_with_boolean_override(self, booly_file):
            pass

        self.addCleanup(delattr, API, "method_with_boolean_override")
        API.method_with_boolean_override = method_with_boolean_override
        self.call_main(["help", "method-with-boolean"], TEST_SCHEMA)
        output = self.stdout.getvalue()
        self.assertIn("Overridden action documentation.", output)
        self.assertIn("booly-file  Booly File", output)
        self.assertIn(
            "other       Parameter documentation exists! (true or false)",
            output)

    def test_bad_scheme(self):
        """
        Specifying a uri without a scheme (or without an http/https scheme)
        results in a nice error.
        """
        _API._run_query = staticmethod(run_query)
        self.call_main(["--uri", "localhost:9090", "get-foo", "1"],
                       TEST_SCHEMA)
        self.assertEqual([1], self.exits)
        self.assertIn(
            "URL must start with 'http://' or 'https://': localhost:9090",
            self.stderr.getvalue())

    def test_pycurl_error(self):
        """
        PyCurl errors are printed nicely to the output if they are raised.
        """
        def run_query(*args, **kwargs):
            raise pycurl.error(481421, "You're bad")
        _API._run_query = run_query
        self.call_main(["get-foo", "1"], TEST_SCHEMA)
        self.assertEqual([1], self.exits)
        self.assertIn("You're bad", self.stderr.getvalue())


class CallTests(TestCase):
    """
    Tests for the L{API.call} method.
    """
    def setUp(self):
        self.queries = []
        self.api = api_factory(TEST_SCHEMA)("u", "a", "s", schema=TEST_SCHEMA)

        def fake_run_query(*args):
            self.queries.append(args)
            return {"some": "result"}

        self.api.run_query = fake_run_query

    def assertCall(self, name, args):
        self.assertEqual([(name, args)], self.queries)

    def test_missing_parameter(self):
        """
        Missing parameters raise a L{TypeError}.
        """
        error = self.assertRaises(TypeError, self.api.call, "GetFoo")
        self.assertIn("Missing parameter id", str(error))

    def test_extra_parameter(self):
        """
        Extra unknown parameters raise a L{TypeError}.
        """
        error = self.assertRaises(TypeError, self.api.call, "GetFoo",
                                  id=5, bar=3)
        self.assertIn("Extra arguments: {'bar': 3}",
                      str(error))

    def test_call(self):
        """
        A successful invocation of L{API.call} invokes L{API.run_query} with
        encoded parameters.
        """
        self.api.call("GetFoo", id=5)
        self.assertCall("GetFoo", {"id": "5"})

    def test_optional(self):
        """
        Optional paremeters are not required.
        """
        self.api.call("MethodWithOptional")
        self.assertCall("MethodWithOptional", {})

    def test_list(self):
        """
        Lists are encoded as a series of indexed keys.
        """
        self.api.call("MethodWithList", listy=[1, 2, 3])
        self.assertCall("MethodWithList",
                        {"listy.1": "1", "listy.2": "2", "listy.3": "3"})

    def test_structure(self):
        """
        Dictionaries are encoded as a series of values keyed by dotted paths.
        """
        self.api.call("MethodWithStructure",
                      struct={"field": "foo", "value": 5})
        self.assertCall("MethodWithStructure",
                        {"struct.field": "foo", "struct.value": "5"})

    def test_boolean(self):
        """
        Booleans are encoded as lowercased strings.
        """
        self.api.call("MethodWithBoolean", booly=True, other=False)
        self.assertCall("MethodWithBoolean",
                        {"booly": "true", "other": "false"})

    def test_unicode(self):
        """
        Unicode is encoded as utf-8.
        """
        uni = u"B\N{HIRAGANA LETTER O}B"
        self.api.call("MethodWithUnicode", uni=uni)
        self.assertCall("MethodWithUnicode", {"uni": uni.encode("utf-8")})

    def test_raw_string(self):
        """
        Raw strings are encoded without any modification.
        """
        raw = u"B\N{HIRAGANA LETTER O}B".encode("utf-8")
        self.api.call("MethodWithRawString", raw=raw)
        self.assertCall("MethodWithRawString", {"raw": raw})

    def test_date(self):
        """
        L{datetime} objects can be encoded to a string format.
        """
        stamp = datetime(2011, 1, 1, 0, 0, 0)
        self.api.call("MethodWithDate", date=stamp)
        self.assertCall("MethodWithDate", {"date": "2011-01-01T00:00:00Z"})

    def test_enum(self):
        """
        Enums are encoded with a passthrough.
        """
        self.api.call("MethodWithEnum", enum="foo")
        self.assertCall("MethodWithEnum", {"enum": "foo"})

    def test_missing_structure_fields(self):
        """
        Missing a required structure field raises a L{TypeError}.
        """
        error = self.assertRaises(TypeError, self.api.call,
                                  "MethodWithStructure", struct={"value": 3})
        self.assertIn("Missing parameter field", str(error))

    def test_optional_structure_fields(self):
        """
        Optional structure fields are not required.
        """
        self.api.call("MethodWithOptionalStructureField", struct={})
        self.assertCall("MethodWithOptionalStructureField", {})

    def test_extra_structure_field(self):
        """
        Extra, unknown structure fields result in a L{TypeError}.
        """
        error = self.assertRaises(TypeError, self.api.call,
                                  "MethodWithStructure",
                                  struct={"field": "a", "value": 3,
                                          "other": 59})
        self.assertIn("Extra arguments: {'other': 59}", str(error))


class ErrorsTests(TestCase):

    def test_build_exceptions(self):
        """
        L{_build_exceptions} searches a schema for errors and generates a
        L{_ErrorsContainer} with all the errors found.
        """
        schema = {
            "Foo": {"2011-08-01": {
                    "doc": "Foo",
                    "errors": [{"code": "Code1"},
                               {"code": "Code2"}],
                    "parameters": []
                    }
                    },
            "Bar": {"2011-08-01": {
                    "doc": "Bar",
                    "errors": [{"code": "Code3"}],
                    "parameters": []
                    }
                    }
            }
        errors = _build_exceptions(schema)
        code1 = errors.lookup_error("Code1Error")
        code2 = errors.lookup_error("Code2Error")
        code3 = errors.lookup_error("Code3Error")

        for error in code1, code2, code3:
            self.assertNotIdentical(error, None, "error wasn't generated")
            self.assertTrue(issubclass(error, APIError))

        self.assertEqual("Code1Error", code1.__name__)
        self.assertEqual("Code2Error", code2.__name__)
        self.assertEqual("Code3Error", code3.__name__)

    def test_error_codes_ending_in_error(self):
        """
        If an error code already ends in "Error", no "Error" is appended to its
        name.
        """
        schema = {"Foo": {"2011-08-01": {"errors": [{"code": "Code1Error"}]}}}
        errors = _build_exceptions(schema)
        error1 = errors.lookup_error("Code1Error")
        self.assertNotIdentical(None, error1)

    def test_module_name(self):
        """
        L{_ErrorsContainer.add_error} munges the exception's module name to
        point to the errors pseudo-module.
        """
        errors = _build_exceptions({})

        class FooError(Exception):
            pass

        errors.add_error("FooError", FooError)
        foo_error = errors.lookup_error("FooError")
        self.assertEqual("landscape_api.base.errors",
                         foo_error.__module__)
